#pragma once

/// Indicates the major version of LibTorch.
#define TORCH_VERSION_MAJOR 2

/// Indicates the minor version of LibTorch.
#define TORCH_VERSION_MINOR 0

/// Indicates the patch version of LibTorch.
#define TORCH_VERSION_PATCH 1

/// Indicates the version of LibTorch.
#define TORCH_VERSION \
  "2.0.1"
