# -*- coding: utf-8 -*-
op_add = '+'
op_sub = '-'
op_mul = '*'
op_div = '/'
op_eq = '='
op_neq = '!='
op_imp = '=>'
op_matching = '⊳'
op_consistency = '~'
op_precision = '⊑'
op_leq = '≤'
op_lt = '<'
op_gt = '>'
op_mod = '%'
